﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnityEngine;
using UnityEngine.Events;
using TMPro;
using System.Reflection;
using System.Globalization;
using System.IO;

namespace Utility01
{
    public static class Utility
    {
        public static string modPath;
        public static string cultureInfo = CultureInfo.CurrentCulture.Name.ToLower();
        public static Vector3 GetModuleVector(this Vector3 vector)
        {
            return new Vector3(Math.Abs(vector.x), Math.Abs(vector.y), Math.Abs(vector.z));
        }
        public static void SetPositionWithOffSet(Transform child, Transform parent, float offSetX = 0, float offSetY = 0)
        {
            child.position = parent.position;
            child.localPosition = new Vector3(offSetX, offSetY);
            child.localRotation = Quaternion.identity;
        }
        public static ItemButtonBehaviour GetItemButtonBehaviour(SpawnableAsset spawnableAsset)
        {
            List<ItemButtonBehaviour> itemButtonBehaviours = (List<ItemButtonBehaviour>)typeof(CatalogBehaviour).GetField("items", BindingFlags.NonPublic | BindingFlags.Instance).GetValue(GameObject.FindObjectOfType<CatalogBehaviour>()); // zooi <3
            return itemButtonBehaviours.Where(itemButton => itemButton.Item == spawnableAsset).FirstOrDefault();
        }
        public static bool IsVisible(this GameObject gameObject)
        {
            if(gameObject.TryGetComponent<SpriteRenderer>(out SpriteRenderer spriteRenderer))
            {
                if (spriteRenderer.isVisible)
                {
                    return true;
                }
            }
            else
            {
                foreach(var spriteRendererChild in gameObject.transform.GetComponentsInChildren<SpriteRenderer>())
                {
                    if (spriteRendererChild.isVisible)
                    {
                        return true;
                    }
                }
            }
            return false;
        }
        public static string GetTranslatedString(string en, string ru)
        {
            return cultureInfo.Contains("ru") ? ru : en;
        }
        public static JointMotor2D SetMotorForce(JointMotor2D jointMotor2D, float force, float torque = 10000f)
        {
            jointMotor2D.motorSpeed = force;
            jointMotor2D.maxMotorTorque = torque;
            return jointMotor2D;
        }
        public static void SetField<T>(this T obj, string nameField, object value)
        {
            Debug.Log(typeof(T).GetFields(BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic).Where(field => field.Name == nameField).FirstOrDefault().Name);
            typeof(T).GetFields(BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic).Where(field => field.Name == nameField).FirstOrDefault().SetValue(obj, value);
        }
        public static object GetField<T>(this T obj, string nameField)
        {
            return typeof(T).GetFields(BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic).Where(field => field.Name == nameField).FirstOrDefault().GetValue(obj);
        }
        public static void DrawCircle(this GameObject gameObject, float radius = 0.3f)
        {
            var pointDebugger = gameObject.AddComponent<PointDebuggerCircle>();
            pointDebugger.radius = radius;
        }
        public static void DrawVector(this Vector2 vector, float radius = 0.3f)
        {
            var circleHandler = new GameObject("colliderHandler");
            var pointBehaviour = circleHandler.AddComponent<PointDebuggerVector>();
            pointBehaviour.radius = radius;
            pointBehaviour.vector = vector;
        }
        public static bool IsColissionDisabled(this GameObject gameObject)
        {
            return gameObject.layer == 10 ? true : false;
        }
        public static void DrawCollider(this GameObject gameObject)
        {
            gameObject.AddComponent<PointDebuggerCollider>().collider = gameObject.GetComponent<Collider2D>();
        }
        public static void DrawCollider(this Collider2D collider)
        {
            var colliderHandler = new GameObject("colliderHandler");
            colliderHandler.AddComponent<PointDebuggerCollider>().collider = collider;
        }
        public static void Destroy(this UnityEngine.Object @object)
        {
            try
            {
                UnityEngine.Object.Destroy(@object);
            }
            catch { }
        }
        public static Vector2 GetDirection(this Vector2 v, Vector2 other)
        {
            Vector2 direction = other - v;

            float x = direction.x == 0 ? 0 : direction.x / (Mathf.Abs(direction.x));
            float y = direction.y == 0 ? 0 : direction.y / (Mathf.Abs(direction.y));

            return new Vector2(x, y);
        }
        public static void IgnoreCollision(this GameObject[] gameObjects, bool ignore = true)
        {
            foreach(var gameObject in gameObjects)
            {
                foreach(var gameObjectToCollision in gameObjects.Where(gameObjectFilter => gameObjectFilter != gameObject))
                {
                    Physics2D.IgnoreCollision(gameObject.GetComponent<Collider2D>(), gameObjectToCollision.GetComponent<Collider2D>(), ignore);
                }
            }
        }
        public static void SetSprite(this GameObject gameObject, Sprite sprite, bool fixColliders = true, string sortingLayerName = "Default", int sortingOrder = 4)
        {
            if (!gameObject.TryGetComponent<SpriteRenderer>(out SpriteRenderer spriteRenderer))
            {
                gameObject.AddComponent<SpriteRenderer>();
            }
            var renderer = gameObject.GetComponent<SpriteRenderer>();
            renderer.sprite = sprite;
            renderer.sortingLayerName = sortingLayerName;
            renderer.sortingOrder = sortingOrder;
            if (fixColliders)
            {
                gameObject.FixColliders();
            }
        }
        public static void ShowDialog(DialogBox dialogBox)
        {
            dialogBox.gameObject.SetActive(true);
        }
        public static void HideDialog(DialogBox dialogBox)
        {
            dialogBox.gameObject.SetActive(false);
        }
        public static Collider2D[] InCollider(this Collider2D collider)
        {
            List<Collider2D> colliders = new List<Collider2D>();
            collider.OverlapCollider(new ContactFilter2D(), colliders);
            return colliders.ToArray();
        }
        public static TMP_FontAsset GetTMP_FontAsset(string nameFont)
        {
            switch (nameFont)
            {
                case "Pixel": return ModAPI.FindSpawnable("Thermometer").Prefab.transform.Find("Text").GetComponent<TextMeshPro>().font;
                case "Clock": return ModAPI.FindSpawnable("Infrared Thermometer").Prefab.transform.Find("value").GetComponent<TextMeshPro>().font;
                default: return GameObject.Find("Canvas").transform.Find("Pre-release build text").GetComponent<TextMeshProUGUI>().font;
            }
        }
        public static GameObject CreateChildObject(string name, Transform parent)
        {
            var gameObject = new GameObject(name);
            gameObject.transform.SetParent(parent);
            gameObject.transform.localScale = new Vector3(1, 1, 1);
            return gameObject;
        }
        public static GameObject CreateChildObject(GameObject prefab, Transform parent)
        {
            var gameObject = GameObject.Instantiate(prefab);
            gameObject.transform.SetParent(parent);
            gameObject.transform.localScale = new Vector3(1, 1, 1);
            return gameObject;
        }
        public static void InitializePhysicalComponent(this GameObject gameObject)
        {
            gameObject.layer = LayerMask.NameToLayer("Objects");
            gameObject.AddComponent<Rigidbody2D>();
            PhysicalBehaviour physicalBehaviour = gameObject.AddComponent<PhysicalBehaviour>();
            physicalBehaviour.Properties = ModAPI.FindPhysicalProperties("Metal");
            physicalBehaviour.SpawnSpawnParticles = false;
            physicalBehaviour.OverrideShotSounds = Array.Empty<AudioClip>();
            physicalBehaviour.OverrideImpactSounds = Array.Empty<AudioClip>();
        }
        public static void OpenLink(string url)
        {
            Type type = Type.GetType("UnityEngine.Application, UnityEngine.CoreModule");
            type.GetMethod("OpenURL", BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static).Invoke(null, new object[] { url });
        }
        public static string GetHierarchyPath(Transform startTransform, Transform endTransform)
        {
            try
            {
                string hierarchyPath = endTransform.name;
                if (endTransform.root == startTransform.root)
                {
                    Transform lastTransform = endTransform;
                    while (true)
                    {
                        if (lastTransform.parent == endTransform.root)
                        {
                            break;
                        }
                        hierarchyPath = lastTransform.parent.name + "/" + hierarchyPath;
                        lastTransform = lastTransform.parent;
                        if (lastTransform == startTransform)
                        {
                            break;
                        }
                    }
                    return hierarchyPath;
                }
                return hierarchyPath;
            }
            catch
            {
                return "";
            }
        }
        public static void ModMetaCache(string modPath)
        {
            Utility.modPath = modPath;
        }
        public static Sprite LoadSprite(string path, bool withoutIncludeModPath = false)
        {
            return Utils.LoadSprite((withoutIncludeModPath == true ? path : modPath + "\\" + path), FilterMode.Point, 35, false);
        }
        public static Texture LoadTexture(string path)
        {
            return Utils.LoadTexture(modPath + "\\" + path);
        }
        public static void ChangeSpecificLimbSprite(this LimbBehaviour limbBehaviour, Sprite skin, Sprite flash, Sprite bone, Sprite damage)
        {
            var limbSpriteRenderer = limbBehaviour.GetComponent<SpriteRenderer>();
            limbSpriteRenderer.sprite = skin;
            limbSpriteRenderer.material.SetTexture("_FleshTex", flash.texture);
            limbSpriteRenderer.material.SetTexture("_BoneTex", bone.texture);
            limbSpriteRenderer.material.SetTexture("_DamageTex", damage.texture);
        }
        public static void ChangeSpecificLimbSprite(this SpriteRenderer spriteRenderer, Sprite skin, Sprite flash, Sprite bone, Sprite damage)
        {
            spriteRenderer.sprite = skin;
            spriteRenderer.material.SetTexture("_FleshTex", flash.texture);
            spriteRenderer.material.SetTexture("_BoneTex", bone.texture);
            spriteRenderer.material.SetTexture("_DamageTex", damage.texture);
        }
        public static GameObject FindPrefab(string name)
        {
            return Resources.Load<GameObject>(name);
        }
        public static void UpdateOutline(this GameObject gameObject)
        {
            if (gameObject.TryGetComponent<PhysicalBehaviour>(out PhysicalBehaviour physicalBehaviour))
            {
                physicalBehaviour.RefreshOutline();
            }
        }
        public static void ErrorNotify(string message)
        {
            ModAPI.Notify($"<color=\"red\">{message}");
        }
        public static void SuccessfulNotify(string message)
        {
            ModAPI.Notify($"<color=\"green\">{message}");
        }
        public static PhysicalProperties GetPhysicalProperties(this GameObject gameObject)
        {
            if (gameObject.TryGetComponent<PhysicalBehaviour>(out PhysicalBehaviour physicalBehaviour))
            {
                return physicalBehaviour.Properties;
            }
            else
            {
                return null;
            }
        }
        public static float GetAverageImpulse(this Collision2D collision)
        {
            var averageImpulse = Utils.GetAverageImpulse(collision.contacts, collision.contacts.Length);
            return averageImpulse;
        }
        public static void CreateFolder(string path)
        {
            Type.GetType("System.IO.Directory").GetMethods().Where(method => method.Name == "CreateDirectory").FirstOrDefault().Invoke(null, new object[] { path });
        }
        public static bool IsFolderExists(string path)
        {
            return (bool)Type.GetType("System.IO.Directory").GetMethods().Where(method => method.Name == "Exists").FirstOrDefault().Invoke(null, new object[] { path });
        }
        public static string[] GetFileSystemEntriesFolder(string path)
        {
            return (string[])Type.GetType("System.IO.Directory").GetMethods().Where(method => method.Name == "GetFileSystemEntries").FirstOrDefault().Invoke(null, new object[] { path });
        }
    }
    public class PointDebuggerCircle : MonoBehaviour
    {
        public float radius = 0.3f;
        private void Update()
        {
            ModAPI.Draw.Circle(gameObject.transform.position, radius);
        }
    }
    public class PointDebuggerVector : MonoBehaviour
    {
        public float radius = 0.3f;
        public Vector2 vector;
        private void Update()
        {
            ModAPI.Draw.Circle(vector, radius);
        }
    }
    public class PointDebuggerCollider : MonoBehaviour
    {
        public Collider2D collider;
        private void Update()
        {
            ModAPI.Draw.Collider(collider);
        }
    }
    public class FreezerDrag : MonoBehaviour
    {
        public FreezeBehaviour freezeBehaviour;
        private void Start()
        {
            OnMouseUp();
        }
        private void OnMouseUp()
        {
            freezeBehaviour = gameObject.GetOrAddComponent<FreezeBehaviour>();
        }
        private void OnMouseDown()
        {
            UnityEngine.Object.Destroy(freezeBehaviour);
        }
    }
    public class PID
    {
        public float Kp, Ki, Kd;

        private float lastError;
        private float P, I, D;

        public PID()
        {
            Kp = 1f;
            Ki = 0;
            Kd = 0.2f;
        }

        public PID(float pFactor, float iFactor, float dFactor)
        {
            this.Kp = pFactor;
            this.Ki = iFactor;
            this.Kd = dFactor;
        }

        public float Update(float error, float dt)
        {
            P = error;
            I += error * dt;
            D = (error - lastError) / dt;
            lastError = error;

            float CO = P * Kp + I * Ki + D * Kd;

            return CO;
        }
    }
    public class ActOnCollideOwn : MonoBehaviour
    {
        public float ImpactForceThreshold;
        public float DispatchChance = 0.5f;
        public Action<Collision2D> Actions;
        public bool debug = false;
        private void OnCollisionEnter2D(Collision2D coll)
        {
            ContactPoint2D[] contacts = coll.contacts;
            var averageImpact = Utils.GetAverageImpulse(contacts, contacts.Length);
            if (debug)
            {
                Debug.Log(averageImpact);
            }
            if (UnityEngine.Random.Range(0f, 1f) > DispatchChance)
            {
                return;
            }
            if (averageImpact > ImpactForceThreshold)
            {
                this.Actions.Invoke(coll);
            }
        }
    }

}